// Copyright (C) 2016   Kire Pûdsje   <kirepudsje@gmail.com>
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, see <http://www.gnu.org/licenses/>.


// heavily based on the work of Ted Burke (batchloaf)
// https://batchloaf.wordpress.com/commandcam/
// and the work of Matjaz Rihtar
// https://github.com/mrihtar/CommandCam/blob/master/CommandCam.cpp

// This is currently a work in progress.
// Very dirty programming
// Error handling terrible
// Argument validation needs to be implemented
// Resolution cannot be set (yet)

#include <octave/oct.h>

// DirectShow header file
#include <dshow.h>
#ifdef __MINGW32__
#include <qedit.h>
#else
// This is a workaround for the missing header
// file qedit.h which seems to be absent from the
// Windows SDK versions 7.0 and 7.1.
// To use the items defined in this dll, the
// DexterLib namespace must be specified.
// The items in question are:
//
//   DexterLib::_AMMediaType
//   DexterLib::ISampleGrabber
//   DexterLib::IID_ISampleGrabber
//
#import "qedit.dll" raw_interfaces_only named_guids
using namespace DexterLib;
#endif

// For some reason, these are not included in the
// DirectShow headers. However, they are exported
// by strmiids.lib, so I'm just declaring them
// here as extern.
EXTERN_C const CLSID CLSID_NullRenderer;
EXTERN_C const CLSID CLSID_SampleGrabber;

// Some common Webcam compression formats
#define sRGB 0x00000000 // RGB888, 24 bits (no compression)

class kpc_webcam_class : public octave_base_value
{
protected:
  IGraphBuilder *pGraph;
  ICaptureGraphBuilder2 *pBuilder;
  IMoniker *pMoniker;
  IPropertyBag *pPropBag;
  IBaseFilter *pCap;
  IBaseFilter *pSampleGrabberFilter;
  ISampleGrabber *pSampleGrabber;
  IBaseFilter *pNullRenderer;
  IMediaControl *pMediaControl;
  int device_idx;
public:
  Cell available_resolutions;
  std::string resolution;

  // Properties
  bool is_constant (void) const { return true;}
  bool is_defined (void) const { return true;}
  bool is_object (void) const { return true;}

  kpc_webcam_class(int devidx = 1);
  ~kpc_webcam_class();
  void construct(int devidx);
  void destruct();
  void obtain_resolutions();
  void set_resolution(const std::string& res);
  uint8NDArray snapshot();
  
  octave_value get_video_property(long prop);
  void set_video_property(long prop, const octave_value& val);
  octave_value get_camera_property(long prop);
  void set_camera_property(long prop, const octave_value& val);
  octave_value get_other_property(long prop);
  void set_other_property(long prop, const octave_value& val);
  octave_value get(const std::string &name);
  void set(const std::string &name, const octave_value& val);
  octave_value get();
  void set(const octave_value_list& args);
};

// Release the format block for a media type
static void _FreeMediaType(AM_MEDIA_TYPE& mt)
{
  if (mt.cbFormat != 0) {
    CoTaskMemFree((PVOID)mt.pbFormat);
    mt.cbFormat = 0;
    mt.pbFormat = NULL;
  }
  if (mt.pUnk != NULL) {
    // pUnk should not be used.
    mt.pUnk->Release();
    mt.pUnk = NULL;
  }
} // _FreeMediaType

// Delete a media type structure that was allocated on the heap
static void _DeleteMediaType(AM_MEDIA_TYPE *pmt)
{
  if (pmt != NULL) {
    _FreeMediaType(*pmt);
    CoTaskMemFree(pmt);
  }
} // _DeleteMediaType

#define STRINGIFY2(x) #x 
#define STRINGIFY(x) STRINGIFY2(x) 
#define pretty_throw(msg) throw(__FILE__ ":" STRINGIFY(__LINE__) " " msg)

static IMoniker *find_moniker(int devidx)
{
  ICreateDevEnum *pDevEnum = NULL;
  IEnumMoniker *pEnum = NULL;
  IMoniker *pMoniker = NULL;
  // Create system device enumerator
  if (CoCreateInstance(CLSID_SystemDeviceEnum, NULL, CLSCTX_INPROC_SERVER,
      IID_PPV_ARGS(&pDevEnum)) != S_OK)
    pretty_throw("Could not create system device enumerator");
  // Video input device enumerator
  HRESULT hr = pDevEnum->CreateClassEnumerator(
      CLSID_VideoInputDeviceCategory, &pEnum, 0);
  if (hr == S_FALSE) {
    pDevEnum->Release();
    pretty_throw("Error enumerating video input devices");
  }
  else if (hr == S_OK) {
    while (pEnum->Next(1, &pMoniker, NULL) == S_OK) {
      if (--devidx == 0)
        break;
      pMoniker->Release();
      pMoniker = NULL;
    }
    pEnum->Release();
  }
  pDevEnum->Release();
  if (devidx)
    return NULL;
  else
    return pMoniker;
}  

kpc_webcam_class::kpc_webcam_class(int devidx) :
    pGraph(), pBuilder(), pMoniker(), pPropBag(), pCap(),
    pSampleGrabberFilter(), pSampleGrabber(), pNullRenderer(),
    pMediaControl()
{
  construct(devidx);
  device_idx = devidx;
}

kpc_webcam_class::~kpc_webcam_class()
{
  // Stop the graph
  pMediaControl->Stop();
  destruct();
}

void kpc_webcam_class::construct(int devidx)
{
  try {
    // Create filter graph
    if (CoCreateInstance(CLSID_FilterGraph, NULL, CLSCTX_INPROC_SERVER,
        IID_IGraphBuilder, (void**)&pGraph) != S_OK)
      pretty_throw("Could not create filter graph");
    // Create capture graph builder
    if (CoCreateInstance(CLSID_CaptureGraphBuilder2, NULL, CLSCTX_INPROC_SERVER,
        IID_ICaptureGraphBuilder2, (void **)&pBuilder) != S_OK)
      pretty_throw("Could not create capture graph builder");
    // Attach capture graph builder to graph
    if (pBuilder->SetFiltergraph(pGraph) != S_OK)
      pretty_throw("Could not attach capture graph builder to graph");      
    
    pMoniker = find_moniker(devidx);
    if (!pMoniker)
      throw("Device not found"); // normal user error
    // Create capture filter and add to graph
    if (pMoniker->BindToObject(0, 0, IID_IBaseFilter, (void**)&pCap) != S_OK)
      pretty_throw("Could not create capture filter");
    // Add capture filter to graph
    if (pGraph->AddFilter(pCap, L"Capture Filter") != S_OK)
      pretty_throw("Could not add capture filter to graph");
    // Create sample grabber filter
    if (CoCreateInstance(CLSID_SampleGrabber, NULL, CLSCTX_INPROC_SERVER,
        IID_IBaseFilter, (void**)&pSampleGrabberFilter) != S_OK)
      pretty_throw("Could not create Sample Grabber filter");
    // Query the ISampleGrabber interface of the sample grabber filter
    if (pSampleGrabberFilter->QueryInterface(IID_ISampleGrabber,
        (void**)&pSampleGrabber) != S_OK)
      pretty_throw("Could not get ISampleGrabber interface to sample grabber filter");
    // Enable sample buffering in the sample grabber filter
    if (pSampleGrabber->SetBufferSamples(true) != S_OK)
      pretty_throw("Could not enable sample buffering in the sample grabber");
    // Set media type in sample grabber filter
    AM_MEDIA_TYPE mt;
    ZeroMemory(&mt, sizeof(AM_MEDIA_TYPE));
    mt.majortype = MEDIATYPE_Video;
    mt.subtype = MEDIASUBTYPE_RGB24;
    if (pSampleGrabber->SetMediaType((_AMMediaType *)&mt) != S_OK)
      pretty_throw("Could not set media type in sample grabber");
    obtain_resolutions();
    if (resolution.empty())
      resolution = available_resolutions(0).string_value();
//    set_resolution(resolution);
    // Add sample grabber filter to filter graph
    if (pGraph->AddFilter(pSampleGrabberFilter, L"SampleGrab") != S_OK)
      pretty_throw("Could not add Sample Grabber to filter graph");
    // Create Null Renderer filter
    if (CoCreateInstance(CLSID_NullRenderer, NULL, CLSCTX_INPROC_SERVER,
        IID_IBaseFilter, (void**)&pNullRenderer) != S_OK)
      pretty_throw("Could not create Null Renderer filter");
    // Add Null Renderer filter to filter graph
    if (pGraph->AddFilter(pNullRenderer, L"NullRender") != S_OK)
      pretty_throw("Could not add Null Renderer to filter graph");
    // Connect up the filter graph's capture stream
    if (pBuilder->RenderStream(&PIN_CATEGORY_CAPTURE, &MEDIATYPE_Video,
        pCap, pSampleGrabberFilter, pNullRenderer) != S_OK)
      pretty_throw("Could not render capture video stream");
    // Get media control interfaces to graph builder object
    if (pGraph->QueryInterface(IID_IMediaControl, (void**)&pMediaControl)
        != S_OK)
      pretty_throw("Could not get media control interface");
    while (1) {
      HRESULT hr = pMediaControl->Run();
      // Hopefully, the return value was S_OK or S_FALSE
      if (hr == S_OK) break; // graph is now running
      if (hr == S_FALSE) continue; // graph still preparing to run
      // If the Run function returned something else,
      // there must be a problem
      //pretty_throw("Could not run filter graph");
      throw("Camera already opened?");
    }
  }
  catch(...) {
    destruct();
    throw;
  }
}

void kpc_webcam_class::destruct() {
  if (pSampleGrabber)
    pSampleGrabber->Release();
  if (pSampleGrabberFilter)
    pSampleGrabberFilter->Release();
  if (pCap)
    pCap->Release();
  if (pPropBag)
    pPropBag->Release();
  if (pMoniker)
    pMoniker->Release();
  if (pBuilder)
    pBuilder->Release();
  if (pGraph)
    pGraph->Release();
}

void kpc_webcam_class::obtain_resolutions()
{
  int nres = 0;
  IEnumPins *pEnumPins = NULL;
  IPin *pPin = NULL;
  IAMStreamConfig *pConfig = NULL;
  AM_MEDIA_TYPE *pmt = NULL;
  BYTE *pSCC = NULL;
  
  available_resolutions.resize(dim_vector(0, 0));
  try {
    if (pCap->EnumPins(&pEnumPins) != S_OK)
      pretty_throw("Could not enumerate pins in base filter");
    while (pEnumPins->Next(1, &pPin, NULL) == S_OK && pPin) {
      PIN_INFO PinInformation;
      if (pPin->QueryPinInfo(&PinInformation) != S_OK)
        pretty_throw("Could not get pin information");
      if (pPin->QueryInterface(IID_IAMStreamConfig, (void**)&pConfig) == S_OK) {
        int count = 0;
        int size = 0;
        if (pConfig->GetNumberOfCapabilities(&count, &size) != S_OK)
          pretty_throw("Could not get number of capabilities");
        try {
          pSCC = new BYTE[size];
        } catch(std::bad_alloc& ba) { 
          throw(ba.what());
        }
        for (int ii = 0; ii < count; ii++) {
          if (pConfig->GetStreamCaps(ii, &pmt, pSCC) != S_OK)
            pretty_throw("Could not get stream capability");
          if (pmt->formattype == FORMAT_VideoInfo) {
            VIDEOINFOHEADER *pVih = (VIDEOINFOHEADER *)pmt->pbFormat;
            available_resolutions.resize(dim_vector(nres + 1, 1));
            available_resolutions(nres++) = std::to_string(pVih->bmiHeader.biWidth) +
                'x' + std::to_string(pVih->bmiHeader.biHeight);
          }
          _DeleteMediaType(pmt);
          pmt = NULL;
        }
        delete[] pSCC;
        pSCC = NULL;
      }
      pPin->Release();
    }
  }
  catch(const char *errmsg) {
    if (pmt)
      _DeleteMediaType(pmt);
    if (pSCC)
      delete[] pSCC;
    if (pPin)
      pPin->Release();
    throw(errmsg);
  }
}

void kpc_webcam_class::set_resolution(const std::string& res)
{
  IEnumPins *pEnumPins = NULL;
  IPin *pPin = NULL;
  IAMStreamConfig *pConfig = NULL;
  AM_MEDIA_TYPE *pmt = NULL;
  BYTE *pSCC = NULL;
  bool formatSet = false;
  
  try {
    if (pCap->EnumPins(&pEnumPins) != S_OK)
      pretty_throw("Could not enumerate pins in base filter");
    while (pEnumPins->Next(1, &pPin, NULL) == S_OK && pPin && !formatSet) {
      PIN_INFO PinInformation;
      if (pPin->QueryPinInfo(&PinInformation) != S_OK)
        pretty_throw("Could not get pin information");
      if (pPin->QueryInterface(IID_IAMStreamConfig, (void**)&pConfig) == S_OK) {
        int count = 0;
        int size = 0;
        if (pConfig->GetNumberOfCapabilities(&count, &size) != S_OK)
          pretty_throw("Could not get number of capabilities");
        pSCC = new BYTE[size];
        for (int ii = 0; ii < count && !formatSet; ii++) {
          if (pConfig->GetStreamCaps(ii, &pmt, pSCC) != S_OK)
            pretty_throw("Could not get stream capability");
          if (pmt->formattype == FORMAT_VideoInfo) {
            VIDEOINFOHEADER *pVih = (VIDEOINFOHEADER *)pmt->pbFormat;
            if (std::to_string(pVih->bmiHeader.biWidth) + 'x' + 
                std::to_string(pVih->bmiHeader.biHeight) == res) {
              if (pConfig->SetFormat(pmt) != S_OK)
                pretty_throw("Could not set media type in sample grabber");
              formatSet = true;   
            }
          }
          _DeleteMediaType(pmt);
          pmt = NULL;
        }
        delete[] pSCC;
        pSCC = NULL;
      }
      pPin->Release();
    }

    // Add sample grabber filter to filter graph
    if (pGraph->AddFilter(pSampleGrabberFilter, L"SampleGrab") != S_OK)
      pretty_throw("Could not add Sample Grabber to filter graph");
    // Create Null Renderer filter
    if (CoCreateInstance(CLSID_NullRenderer, NULL, CLSCTX_INPROC_SERVER,
        IID_IBaseFilter, (void**)&pNullRenderer) != S_OK)
      pretty_throw("Could not create Null Renderer filter");
    // Add Null Renderer filter to filter graph
    if (pGraph->AddFilter(pNullRenderer, L"NullRender") != S_OK)
      pretty_throw("Could not add Null Renderer to filter graph");
    // Connect up the filter graph's capture stream
    if (pBuilder->RenderStream(&PIN_CATEGORY_CAPTURE, &MEDIATYPE_Video,
        pCap, pSampleGrabberFilter, pNullRenderer) != S_OK)
      pretty_throw("Could not render capture video stream");
    // Get media control interfaces to graph builder object
    if (pGraph->QueryInterface(IID_IMediaControl, (void**)&pMediaControl)
        != S_OK)
      pretty_throw("Could not get media control interface");
    while (1) {
      HRESULT hr = pMediaControl->Run();
      // Hopefully, the return value was S_OK or S_FALSE
      if (hr == S_OK) break; // graph is now running
      if (hr == S_FALSE) continue; // graph still preparing to run
      // If the Run function returned something else,
      // there must be a problem
      //pretty_throw("Could not run filter graph");
      throw("Camera already opened?");
    }
  }
  catch(std::bad_alloc& ba) { 
    if (pConfig)
      pConfig->Release();
    if (pPin)
      pPin->Release();
    if (pEnumPins)
      pEnumPins->Release();
    throw(ba.what());
  }
  catch(...) {
    if (pmt)
      _DeleteMediaType(pmt);
    if (pSCC)
      delete[] pSCC;
    if (pConfig)
      pConfig->Release();
    if (pPin)
      pPin->Release();
    if (pEnumPins)
      pEnumPins->Release();
    throw;
  }
  if (!formatSet)
    throw("Could not set desired format"); // normal user error
}

uint8NDArray kpc_webcam_class::snapshot()
{
  uint8NDArray retval;
  uint8_t *imgBuffer = NULL;      
  AM_MEDIA_TYPE mt;
  bool mt_set = false;
  try {
    // Grab a sample
    // First, find the required buffer size
    long imgBufferSize = 0;
    while (1) {
      // Passing in a NULL pointer signals that we're just checking
      // the required buffer size; not looking for actual data yet.
      HRESULT hr = pSampleGrabber->GetCurrentBuffer(&imgBufferSize, NULL);
      // Keep trying until buffer size is set to non-zero value
      if (hr == S_OK && imgBufferSize != 0)
        break;
      // If the return value isn't S_OK or VFW_E_WRONG_STATE
      // then something has gone wrong. VFW_E_WRONG_STATE just
      // means that the filter graph is still starting up and
      // no data has arrived yet in the sample grabber filter.
      if (hr != S_OK && hr != VFW_E_WRONG_STATE)
        pretty_throw("Could not get buffer size");
    }
    // Allocate buffer for image
    imgBuffer = new uint8_t[imgBufferSize];
    // Retrieve image data from sample grabber buffer
    if (pSampleGrabber->GetCurrentBuffer(&imgBufferSize, (long*) imgBuffer) != S_OK)
      pretty_throw("Could not get buffer data from sample grabber");
    // Get the media type from the sample grabber filter
    if (pSampleGrabber->GetConnectedMediaType((_AMMediaType *)&mt) != S_OK)
      pretty_throw("Could not get media type");
    mt_set = true;
    // Retrieve format information
    VIDEOINFOHEADER *pVih = NULL;
    if ((mt.formattype == FORMAT_VideoInfo) &&
        (mt.cbFormat >= sizeof(VIDEOINFOHEADER)) &&
        (mt.pbFormat != NULL)) {
      // Get video info header structure from media type
      pVih = (VIDEOINFOHEADER*)mt.pbFormat;
      if (pVih->bmiHeader.biCompression != sRGB)
        pretty_throw("Expected sRGB compression format");          
      // Convert image to octave format
      retval.resize(dim_vector(pVih->bmiHeader.biHeight,
          pVih->bmiHeader.biWidth, 3));
      uint8_t *src = imgBuffer;
      for (int y = pVih->bmiHeader.biHeight; y--;)
        for (int x = 0; x < pVih->bmiHeader.biWidth; x++)
          for (int c = 3; c--;)
            retval(y,x,c) = *(src++);
    }
    else
      pretty_throw("Grabbed frame has wrong format");
    _FreeMediaType(mt);
    delete[] imgBuffer;
  }
  catch(std::bad_alloc& ba) { 
    throw(ba.what());
  }
  catch(...) {
    if (mt_set)
      _FreeMediaType(mt);
    if (imgBuffer)
      delete[] imgBuffer;
    throw;
  }
  return retval;
}

static std::string wchar_to_utf8(const wchar_t *ws)
{
  std::string retval;
  if (ws) {
    int len = WideCharToMultiByte(CP_UTF8, 0, ws, -1, NULL, 0, NULL, NULL);
    retval.resize(len);
    WideCharToMultiByte(CP_UTF8, 0, ws, -1, &retval[0], len, NULL, NULL);
  }
  return retval;
}

static Cell list_devices()
{
  Cell retval;
  IGraphBuilder *pGraph = NULL;
  ICaptureGraphBuilder2 *pBuilder = NULL;
  ICreateDevEnum *pDevEnum = NULL;
  IMoniker *pMoniker = NULL;
  IEnumMoniker *pEnum = NULL;
  IPropertyBag *pPropBag = NULL;

  try {
    // Create filter graph
    if (CoCreateInstance(CLSID_FilterGraph, NULL, CLSCTX_INPROC_SERVER,
        IID_IGraphBuilder, (void**)&pGraph) != S_OK)
      pretty_throw("Could not create filter graph");
    // Create capture graph builder
    if (CoCreateInstance(CLSID_CaptureGraphBuilder2, NULL, CLSCTX_INPROC_SERVER,
        IID_ICaptureGraphBuilder2, (void **)&pBuilder) != S_OK)
      pretty_throw("Could not create capture graph builder");
    // Attach capture graph builder to graph
    if (pBuilder->SetFiltergraph(pGraph) != S_OK)
      pretty_throw("Could not attach capture graph builder to graph");
    // Create system device enumerator
    if (CoCreateInstance(CLSID_SystemDeviceEnum, NULL, CLSCTX_INPROC_SERVER,
        IID_PPV_ARGS(&pDevEnum)) != S_OK)
      pretty_throw("Could not create system device enumerator");
    // Video input device enumerator
    HRESULT hr = pDevEnum->CreateClassEnumerator(
        CLSID_VideoInputDeviceCategory, &pEnum, 0);
    if (hr != S_OK && hr != S_FALSE)
      pretty_throw("Error enumerating video input devices");
    int n = 0;
    while (pEnum && pEnum->Next(1, &pMoniker, NULL) == S_OK) {
      // Get device name
      if (pMoniker->BindToStorage(0, 0, IID_PPV_ARGS(&pPropBag)) != S_OK)
        pretty_throw("Cannot bind to storage");
      VARIANT var;
      VariantInit(&var);
      pPropBag->Read(L"FriendlyName", &var, NULL);
      retval.resize(dim_vector(n + 1, 1));
      retval(n++) = wchar_to_utf8((wchar_t*) var.bstrVal);
      VariantClear(&var);
      pPropBag->Release();
      pMoniker->Release();
    }
    pEnum->Release();
    pDevEnum->Release();
    pBuilder->Release();
    pGraph->Release();
  }
  catch(const char* errmsg) {
    if (pPropBag)
      pPropBag->Release();
    if (pMoniker)
      pMoniker->Release();
    if (pEnum)
      pEnum->Release();
    if (pDevEnum)
      pDevEnum->Release();
    if (pBuilder)
      pBuilder->Release();
    if (pGraph)
      pGraph->Release();
    throw(errmsg);
  }
  return retval;
}

static std::string get_device_name(IMoniker *moniker)
{
  std::string retval;
  IPropertyBag *propbag = NULL;
  if (moniker->BindToStorage(0, 0, IID_PPV_ARGS(&propbag)) != S_OK)
    error("%s, %s: Cannot bind to storage", __FILE__, __PRETTY_FUNCTION__);
  VARIANT var;
  VariantInit(&var);
  propbag->Read(L"FriendlyName", &var, NULL);
  retval = wchar_to_utf8((wchar_t*) var.bstrVal);
  VariantClear(&var);
  propbag->Release();
  return retval;
}

typedef enum {ISVID, ISCAM, ISOTHER} enumproptype;
typedef enum {PropName, PropRes, PropAvRes} enumotherprop;
static const struct {
  enumproptype proptype;
  bool writable;
  int val;
  const char* name;
} props_lookup[] = {
  { ISOTHER, false, PropName, "Name" },
  { ISOTHER, true, PropRes, "Resolution" },
  { ISOTHER, false, PropAvRes, "AvailableResolutions" },
  { ISVID, true, VideoProcAmp_Brightness, "Brightness" },
  { ISVID, true, VideoProcAmp_Contrast, "Contrast" },
  { ISVID, true, VideoProcAmp_Hue, "Hue" },
  { ISVID, true, VideoProcAmp_Saturation, "Saturation" },
  { ISVID, true, VideoProcAmp_Sharpness, "Sharpness" },
  { ISVID, true, VideoProcAmp_Gamma, "Gamma" },
  { ISVID, true, VideoProcAmp_ColorEnable, "ColorEnable" },
  { ISVID, true, VideoProcAmp_WhiteBalance, "WhiteBalance" },
  { ISVID, true, VideoProcAmp_BacklightCompensation, "BacklightCompensation" },
  { ISVID, true, VideoProcAmp_Gain, "Gain" },
  { ISCAM, true, CameraControl_Pan, "Pan" },
  { ISCAM, true, CameraControl_Tilt, "Tilt" },
  { ISCAM, true, CameraControl_Roll, "Roll" },
  { ISCAM, true, CameraControl_Zoom, "Zoom" },
  { ISCAM, true, CameraControl_Exposure, "Exposure" },
  { ISCAM, true, CameraControl_Iris, "Iris" },
  { ISCAM, true, CameraControl_Focus, "Focus" }
};

  // returns empty string if name does not exist
  octave_value kpc_webcam_class::get_video_property(long prop)
  {
    octave_value retval;
    IAMVideoProcAmp *pProcAmp = NULL;      
    if (pCap->QueryInterface(IID_IAMVideoProcAmp, (void**)&pProcAmp) == S_OK) {
      long intval, autoval;
      if (pProcAmp->Get(prop, &intval, &autoval) == S_OK) {
        if (autoval == VideoProcAmp_Flags_Manual) {
          if (prop == VideoProcAmp_ColorEnable 
              || prop == VideoProcAmp_BacklightCompensation)
            retval = intval ? "on" : "off";
          else
            retval = intval;
        }
        else
          retval = "auto";
      }
    }
    return retval;
  }

  void kpc_webcam_class::set_video_property(long prop, const octave_value& val)
  {
    IAMVideoProcAmp *pProcAmp = NULL;      
    if (pCap->QueryInterface(IID_IAMVideoProcAmp, (void**)&pProcAmp) == S_OK) {
      long intval = 0, autoval = VideoProcAmp_Flags_Manual;
      if (val.is_string()) {
        if (val.string_value() == "auto")
          autoval = VideoProcAmp_Flags_Auto;
        else if (prop == VideoProcAmp_ColorEnable 
            || prop == VideoProcAmp_BacklightCompensation) {      
          if (val.string_value() == "on")
            intval = 1;
          else if (val.string_value() == "off")
            intval = 0;
          else
            error("set: only valid string value for %s property are 'on', 'off' and 'auto'");
        }
        else
          error("set: only valid string value for %s property is 'auto'");
      }
      else {
        intval = val.int_value();
        if (error_state) {
          error("set: expecting scalar value or 'auto' for %s property");
          return;
        }
      }
      pProcAmp->Set(prop, intval, autoval);
    }
  }

  // returns empty string if name does not exist
  octave_value kpc_webcam_class::get_camera_property(long prop)
  {
    octave_value retval;
    IAMCameraControl *pProcAmp = NULL;      
    if (pCap->QueryInterface(IID_IAMCameraControl, (void**)&pProcAmp) == S_OK) {
      long intval, autoval;
      if (pProcAmp->Get(prop, &intval, &autoval) == S_OK) {
        if (autoval == CameraControl_Flags_Manual)
          retval = intval;
        else
          retval = "auto";
      }
    }
    return retval;
  }

  void kpc_webcam_class::set_camera_property(long prop, const octave_value& val)
  {
    IAMCameraControl *pProcAmp = NULL;      
    if (pCap->QueryInterface(IID_IAMCameraControl, (void**)&pProcAmp) == S_OK) {
      long intval = 0, autoval = CameraControl_Flags_Manual;
      if (val.is_string()) {
        if (val.string_value() == "auto")
          autoval = CameraControl_Flags_Auto;
        else
          error("set: only valid string value for %s property is 'auto'");
      }
      else {
        intval = val.int_value();
        if (error_state) {
          error("set: expecting scalar value or 'auto' for %s property");
          return;
        }
      }
      pProcAmp->Set(prop, intval, autoval);
    }
  }

  // returns empty string if name does not exist
  octave_value kpc_webcam_class::get_other_property(long prop)
  {
    switch (prop) {
    case PropName:
      return get_device_name(pMoniker);
    case PropRes:
      return resolution;
    case PropAvRes:
      return available_resolutions;
    default:
      return octave_value();
    }
  }

  // returns empty string if name does not exist
  void kpc_webcam_class::set_other_property(long prop, const octave_value& val)
  {
    switch (prop) {
    case PropRes:
      warning("set: Resolution not implemented (yet)");
  //    destruct();
  //    resolution = val.string_value();
  //    construct(device_idx);
      break;
    }
  }

  octave_value kpc_webcam_class::get(const std::string &name)
  {
    for (unsigned i = 0; i < sizeof(props_lookup) / sizeof(props_lookup[0]); i++) {
      if (name == props_lookup[i].name) {
        switch (props_lookup[i].proptype) {
        case ISVID:
          return get_video_property(props_lookup[i].val);
        case ISCAM:
          return get_camera_property(props_lookup[i].val);
        case ISOTHER:
          return get_other_property(props_lookup[i].val);
        }
      }
    }
    error("get: invalid property %s", name.c_str());
    return octave_value();
  }

  void kpc_webcam_class::set(const std::string &name, const octave_value& val)
  {
    for (unsigned i = 0; i < sizeof(props_lookup) / sizeof(props_lookup[0]); i++) {
      if (name == props_lookup[i].name) {
        if (props_lookup[i].writable) {
          switch (props_lookup[i].proptype) {
          case ISVID:
            set_video_property(props_lookup[i].val, val);
            return;
          case ISCAM:
            set_camera_property(props_lookup[i].val, val);
            return;
          case ISOTHER:
            set_other_property(props_lookup[i].val, val);
            return;
          }
        }
        else {
          error("set: property %s is read-only", name.c_str());          
        }
      }
    }
    error("set: invalid property %s", name.c_str());
  }

  octave_value kpc_webcam_class::get()
  {
    octave_scalar_map retval;
    octave_value val;
    for (unsigned i = 0; i < sizeof(props_lookup) / sizeof(props_lookup[0]); i++) {
      switch (props_lookup[i].proptype) {
      case ISVID:
        val = get_video_property(props_lookup[i].val);
        break;
      case ISCAM:
        val = get_camera_property(props_lookup[i].val);
        break;
      case ISOTHER:
        val = get_other_property(props_lookup[i].val);
        break;
      }
      if (!val.is_empty())
        retval.assign(props_lookup[i].name, val);
    }
    return retval;
  }

  void kpc_webcam_class::set(const octave_value_list& args)
  {
    for (int i = 1; i < args.length(); i += 2) {
      if (args(i).is_string())
        set(args(i).string_value(), args(i + 1));
      else
        error("set: expected property name to be a string");
    }
  }


// PKG_ADD: autoload ("kpc_webcam_get", which ("kpc_webcam"));
// PKG_DEL: try; autoload ("kpc_webcam_get", which ("kpc_webcam"), "remove"); catch; end;
DEFUN_DLD(kpc_webcam_get, args, nargout,
          "-*- texinfo -*-\n\
@deftypefn {} {@var{retval} =} kpc_webcam_get (@var{input1}, @var{input2})\n\
@seealso{}\n\
@end deftypefn")
{
  octave_value_list retval;
  int nargin = args.length ();

  kpc_webcam_class* cam = NULL;
  const octave_base_value& rep = args(0).get_rep();
  cam = &((kpc_webcam_class &)rep);

  if (nargin == 1)
    retval(0) = cam->get();  
  else
    retval(0) = cam->get(args(1).string_value());  
 
  return retval;
}


// PKG_ADD: autoload ("kpc_webcam_set", which ("kpc_webcam"));
// PKG_DEL: try; autoload ("kpc_webcam_set", which ("kpc_webcam"), "remove"); catch; end;
DEFUN_DLD(kpc_webcam_set, args, nargout,
          "-*- texinfo -*-\n\
@deftypefn {} {@var{retval} =} kpc_webcam_get (@var{input1}, @var{input2})\n\
@seealso{}\n\
@end deftypefn")
{
  octave_value_list retval;
  int nargin = args.length ();

  kpc_webcam_class* cam = NULL;
  const octave_base_value& rep = args(0).get_rep();
  cam = &((kpc_webcam_class &)rep);

  if (nargin < 3 || !(nargin & 1))
;//    retval(0) = cam->get();  
  else
    cam->set(args);  
 
  return retval;
}

// PKG_ADD: autoload ("kpc_snapshot", which ("kpc_webcam"));
// PKG_DEL: try; autoload ("kpc_snapshot", which ("kpc_webcam"), "remove"); catch; end;
DEFUN_DLD(kpc_snapshot, args, nargout,
          "-*- texinfo -*-\n\
@deftypefn {} {@var{retval} =} snapshot (@var{input1}, @var{input2})\n\
@seealso{}\n\
@end deftypefn")
{
  octave_value_list retval;

  if (args.length() != 1) {
    print_usage();
    return octave_value();
  }

  kpc_webcam_class* cam = NULL;
  const octave_base_value& rep = args(0).get_rep();
  cam = &((kpc_webcam_class &)rep);
  retval(0) = cam->snapshot();

  return retval;
}


static bool type_loaded = false;


// PKG_ADD: autoload ("kpc_webcam", which ("kpc_webcam"));
// PKG_DEL: try; autoload ("kpc_webcam", which ("kpc_webcam"), "remove"); catch; end;
DEFUN_DLD (kpc_webcam, args, nargout, 
        "-*- texinfo -*-\n\
")
{
  if (!type_loaded)
  {
      kpc_webcam_class::register_type();
      type_loaded = true;
  }
  warning("This is work in progress, be prepared for segfaults");

  int camidx = 1;
  if (args.length() == 0)
    camidx = 1;
  else if (args.length() > 1) {
    print_usage();
    return octave_value();
  }
  else if (args(0).is_string ()) {
    std::string devname = args(0).string_value();
    Cell list = list_devices();
    bool found = false;
    // first compare full string
    for (int i = 0; !found && i < list.numel(); i++) {
      if (list(i).string_value() == devname) {
        camidx = i + 1;
        found = true;
      }
    }
    // then compare first part of string
    for (int i = 0; !found && i < list.numel(); i++) {
      std::string cmp = list(i).string_value();
      if (cmp.length() > devname.length()
          && !cmp.compare(0, devname.length(), devname)) {
        camidx = i + 1;
        found = true;
      }
    }
    if (!found)
      error("Webcam %s not found", devname.c_str());
  }
  else {
    camidx = args(0).int_value();
    if (error_state)
      error("Value should be integer or string");
  }
    
  kpc_webcam_class* cam;
  try {
    cam = new kpc_webcam_class(camidx);
  }
  catch(const char *errmsg) {
    error(errmsg);
  }

	octave_value_list retval;
	retval(0) = cam;
  return retval;
}

// PKG_ADD: autoload ("webcamlist", which ("kpc_webcam"));
// PKG_DEL: try; autoload ("webcamlist", which ("kpc_webcam"), "remove"); catch; end;
DEFUN_DLD(webcamlist, args, nargout,
          "-*- texinfo -*-\n\
@deftypefn {} {@var{retval} =} webcamlist\n\
\n\
List available webcams\n\
\n\
Returns a cell array containg all available UVC webcams\n\
\n\
@seealso{webcam}\n\
@end deftypefn")
{
  octave_value_list retval;
  int nargin = args.length();

  if (nargin || nargout > 1) {
    print_usage();
    return octave_value();
  }

  try {
    retval(0) = list_devices();
  }
  catch(const char* errmsg) {
    error(errmsg);
  }
 
  return retval;
}
